/*
        radino CC1101 UART-Bridge
        for more information: www.in-circuit.de or www.radino.cc

	Copyright (c) 2014 In-Circuit GmbH
        
        v0.1 - 2014.12.19
	    Basic implementation - functions and comments are subject to change
        
	Permission is hereby granted, free of charge, to any person obtaining a copy of this software
	and associated documentation files (the "Software"), to deal in the Software without restriction,
	including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
	LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
	WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <SoftwareSerial.h>
#include <avr/wdt.h>
#include <EEPROM.h>
#include <IC_CC1101.h>

/******************************** Configuration *************************************/

/* Network settings can be modified on the Bridged UART while pulling down HWB
Keycommands:
'/','*': navigate Entries
'+','-': modify current Entry
"nXX": set network ID to XX (as Hex)
"mXX": set my ID to XX (as Hex)
"tXX": set target ID to XX (as Hex)
"hXX": set hopcount to XX (as Hex)
"rXX": set maximum retries to XX (as Hex)
"l": set network ID to XX (as Hex)
"s": save values to eeprom
*/

#define EEPROM_MYSETTINGS_ADDRESS 128 //Starting byte of settings struct in EEPROM
#define EEPROM_CHECKPATTERN 0xA5
struct networkSettingsStruct {
  unsigned char nwid = 0;
  unsigned char myid = 0xFF;
  unsigned char tgtid = 0xFF;
  unsigned char hopc = 1;
  unsigned char retries = 2;
  unsigned char accBC = 1;
  unsigned char chk = EEPROM_CHECKPATTERN;  //let it be the last value -> for checking if ever written settings
} mySettings;

#define RECEIVE_TO 1000  //Wait at max this long in ms for packet arrival


#define DINRAILADAPTER   0  //Is a DinRailAdapter (Bridge SoftwareSerial on pins 10,11,12)
#define RADINOSPIDER     1  //Is a radino Spider (Bridge Serial1)
#define RADINOUARTBRIDGE 0  //Is a radino UART Bridge (Bridge Serial1)
#define RADINOLEONARDO   0  //Is a radino Leonardo (Bridge USB Serial)
#define RADINOUSBBRIDGE  0  //Is a radino USB Bridge (Bridge USB Serial)
#define RJ45_UARTBRIDGE  0  //Is a RJ45-UART Bridge (Bridge Softwareserial on 10,A5,12)


#define UARTBUF_SIZE 256  //Buffer for UART in
#define UART_SENDTHRES 128  //Start sending when this full
#define RADIOBUF_SIZE 256  //Buffer for radio in
#define RADIO_SENDTHRES 128  //Start writing when this full

//delay between packets:
//modify according to expected serial speed and partner reaction speed
//(prevent bufferoverflow, but allow other side to empty receive buffer)
//send after receive: PACKET_GAP
//send after non acked receive: PACKET_GAP+PACKET_PAUSE
//send after sent self: RESEND_TIMEOUT (Is overwritten on ACK receive with PACKET_GAP)
//send after sendError: PACKET_GAP_ERR
//send after sent with noAck: PACKET_GAP_NOACK
#define UART_TIMEOUT 10  //UART is considered empty after this time in ms
#define RESEND_TIMEOUT 75  //Wait this time in ms for an radio ack
#define PACKET_GAP 10 //Delay between receive and outgoing ack
#define PACKET_PAUSE 10 //Delay for ACK when buffer nearly full (Added to GAP)
#define PACKET_GAP_NOACK 50 //Delay between packets when no ack is expected
#define PACKET_GAP_ERR 25 //Delay on failed transmit

#define SEND_RETRIES_ON_ERROR 3 //How often we try to transmit after transmission start failed (e.g. channel busy)

//Used for error signaling (ON when restransmitting, OFF on receive success)
#define ERRLEDON() digitalWrite(13,HIGH)
#define ERRLEDOFF() digitalWrite(13,LOW)
#define ERRLEDINIT() pinMode(13, OUTPUT)

//Write messages on dbgSerial
//0:none
//1:Errors
//2:Information
#define DBGLVL 0

/***************************** Debug messages ***********************/

//Write out errors on dbgSerial
#if DBGLVL>=1
  #define DBGERR(x) dbgSerial.print(x)
  #define DBGERR2(x,y) dbgSerial.print(x,y)
  #define DBGERRWRITE(x) dbgSerial.write(x)
#else
  #define DBGERR(x) do{}while(0)
  #define DBGERR2(x,y) do{}while(0)
  #define DBGERRWRITE(x) do{}while(0)
#endif

//Write out information on dbgSerial
#if DBGLVL>=2
  #define DBGINFO(x) dbgSerial.print(x)
  #define DBGINFO2(x,y) dbgSerial.print(x,y)
  #define DBGWRITE(x) dbgSerial.write(x)
#else
  #define DBGINFO(x) do{}while(0)
  #define DBGINFO2(x,y) do{}while(0)
  #define DBGWRITE(x) do{}while(0)
#endif

/******************************* Module specific settings **********************/
#undef TXEN_PIN

#if DINRAILADAPTER  //DinRailAdapter
  #define dbgSerial Serial
  #define dbgSerialSpeed 57600
  
  #define RX_PIN 10
  #define TX_PIN 11
  SoftwareSerial bridge = SoftwareSerial(RX_PIN,TX_PIN);
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay 0 //SofwareSerial writes blocking - no wait necessary
  #define TXEN_PIN 12  //Which pin to use for RS485 functionality - comment to disable
  
#elif RADINOSPIDER  //radino Spider
  #define dbgSerial Serial
  #define dbgSerialSpeed 57600
  
  #define bridge Serial1
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay ((bridgeBurst*1000)/(bridgeSpeed/8))  //Time between calls - Let hardware serial write out async
//  #define TXEN_PIN 2  //Which pin to use for RS485 functionality - comment to disable

#elif RADINOUARTBRIDGE  //radino UART bridge
  #define dbgSerial Serial
  #define dbgSerialSpeed 57600
  
  #define bridge Serial1
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay ((bridgeBurst*1000)/(bridgeSpeed/8))  //Time between calls - Let hardware serial write out async
//  #define TXEN_PIN 2  //Which pin to use for RS485 functionality - comment to disable
  
#elif RADINOLEONARDO  //radino Leonardo
  #define dbgSerial Serial1
  #define dbgSerialSpeed 57600
  
  #define bridge Serial
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay ((bridgeBurst*1000)/(bridgeSpeed/8))  //Time between calls - Let hardware serial write out async
//  #define TXEN_PIN 2  //Which pin to use for RS485 functionality - comment to disable

#elif RADINOUSBBRIDGE  //radino USB Bridge
  #define dbgSerial Serial1
  #define dbgSerialSpeed 57600
  
  #define bridge Serial
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay ((bridgeBurst*1000)/(bridgeSpeed/8))  //Time between calls - Let hardware serial write out async
//  #define TXEN_PIN 2  //Which pin to use for RS485 functionality - comment to disable

#elif RJ45_UARTBRIDGE  //RJ45 Uartbridge
  #define dbgSerial Serial
  #define dbgSerialSpeed 57600
  
  #define RX_PIN 10
  #define TX_PIN A5
  SoftwareSerial bridge = SoftwareSerial(RX_PIN,TX_PIN);
  #define bridgeSpeed 9600
  #define bridgeBurst 10  //Chars written out per call
  #define bridgeDelay 0 //SofwareSerial writes blocking - no wait necessary
  #define TXEN_PIN 12  //Which pin to use for RS485 functionality - comment to disable

#endif  //Module selection

//Utilize RS485 functionality?
#ifdef TXEN_PIN
  #define TXEN_ON() digitalWrite(TXEN_PIN, HIGH)
  #define TXEN_OFF() digitalWrite(TXEN_PIN, LOW)
  #define TXEN_INIT() pinMode(TXEN_PIN, OUTPUT)
#else
  #define TXEN_ON() do{}while(0)
  #define TXEN_OFF() do{}while(0)
  #define TXEN_INIT() do{}while(0)
#endif

/****************** Packet structures *******************/
#define PAKETSIZE 61  //CC1101 adds LEN, LQI, RSSI -- stay under fifo size of 64 byte (CC1101 buggy)

//Header for our data packets
typedef struct
{
  uint8_t nwid;
  uint8_t src;
  uint8_t dest;
  uint8_t seq;
  uint8_t pseq;
  uint8_t hopc;
  uint8_t len;
  uint8_t crc;
} header_t;

#define MAXDATALEN (PAKETSIZE-sizeof(header_t))

//Structure of our data packets
typedef struct
{
  header_t header;
  uint8_t data[MAXDATALEN];
} packet_t;

//Packet format delivered by the CC1101 RX
typedef struct
{
  uint8_t len;
  packet_t packet;
  uint16_t appended;
} transfer_t;

/****************** Vars *******************************/
transfer_t RX_buffer = {0,};
packet_t TX_packet = {0,};

char uartBuf[UARTBUF_SIZE] = {0,};
unsigned short uartBufWrite = 0;
unsigned short uartBufRead = 0;
#define uartBuf_Len() ((uartBufWrite+UARTBUF_SIZE-uartBufRead)%UARTBUF_SIZE)
#define uartBuf_available() uartBufLen()
#define uartBuf_NE() (uartBufWrite!=uartBufRead?1:0)
#define uartBuf_NF()	(uartBuf_NextWrite()!=uartBufRead)
#define uartBuf_Clear() (uartBufRead=uartBufWrite)
#define uartBuf_NextRead() ((uartBufRead+1)%UARTBUF_SIZE)
#define uartBuf_NextWrite() ((uartBufWrite+1)%UARTBUF_SIZE)
#define uartBuf_Free() (UARTBUF_SIZE-uartBuf_Len())
uint8_t uartBuf_read()
{
  uint8_t ret = '\0';
  if (uartBuf_NE())
  {
    ret = uartBuf[uartBufRead];
    uartBufRead = uartBuf_NextRead();
  }
  return ret;
}
uint8_t uartBuf_write(uint8_t chr)
{
  if (uartBuf_NF())
  {
    uartBuf[uartBufWrite] = chr;
    uartBufWrite = uartBuf_NextWrite();
    return 1;
  }
  return 0;
}
uint16_t uartBuf_print(String str)
{
  uint16_t i=0;
  while(str[i]!='\0')
  {
    if (!uartBuf_write(str[i++])) return 0;
  }
  return i;
}

char radioBuf[RADIOBUF_SIZE] = {0,};
unsigned short radioBufWrite = 0;
unsigned short radioBufRead = 0;
#define radioBuf_Len() ((radioBufWrite+RADIOBUF_SIZE-radioBufRead)%RADIOBUF_SIZE)
#define radioBuf_available() radioBufLen()
#define radioBuf_NE() (radioBufWrite!=radioBufRead?1:0)
#define radioBuf_NF() (radioBuf_NextWrite()!=radioBufRead)
#define radioBuf_Clear() (radioBufRead=radioBufWrite)
#define radioBuf_NextRead() ((radioBufRead+1)%RADIOBUF_SIZE)
#define radioBuf_NextWrite() ((radioBufWrite+1)%RADIOBUF_SIZE)
#define radioBuf_Free() (RADIOBUF_SIZE-radioBuf_Len())
uint8_t radioBuf_read()
{
  uint8_t ret = '\0';
  if (radioBuf_NE())
  {
    ret = radioBuf[radioBufRead];
    radioBufRead = radioBuf_NextRead();
  }
  return ret;
}
uint8_t radioBuf_write(uint8_t chr)
{
  if (radioBuf_NF())
  {
    radioBuf[radioBufWrite] = chr;
    radioBufWrite = radioBuf_NextWrite();
    return 1;
  }
  return 0;
}
uint16_t radioBuf_print(String str)
{
  uint16_t i=0;
  while(str[i]!='\0')
  {
    if (!radioBuf_write(str[i++])) return 0;
  }
  return i;
}

//Convert hex char to uint
unsigned char fromhexch(unsigned char ch)
{
  if (ch>='0' && ch<='9')  return (ch-'0');
  if (ch>='a' && ch<='f')  return (ch-'a'+10);
  if (ch>='A' && ch<='F')  return (ch-'A'+10);
  return 0;
}

IC_CC1101 wireless;  //The CC1101 device

/************************* Functions **********************/

//Initialize the system
void setup()
{
  dbgSerial.begin(dbgSerialSpeed);
  dbgSerial.print("\r\n\r\nHello\r\n\r\n");
  wdt_enable(WDTO_8S);  //Watchdog 8s
  
  //Init bridged serial and TXE for RS485 if required
  bridge.begin(bridgeSpeed);
  TXEN_INIT();
  
  //Init LED13 for showing errors
  ERRLEDINIT(); ERRLEDOFF();
  
  //Initialize wireless chip
  wireless.Init(PAKETSIZE,RECEIVE_TO);
  
  //Use HWB for configuration via bridged serial
  DDRE &= ~(0x04);  //HWB as input
  PORTE |= 0x04;    //Enable PU
#define HWB() (PINE&0x04)
  
  //Load settings from EEPROM. If invalid -> set defaults
  if ( EEPROM_CHECKPATTERN!=EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+(sizeof(struct networkSettingsStruct))-1) )
    for(unsigned char i=0;i<sizeof(struct networkSettingsStruct);i++)
      if (EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+i)!=(((unsigned char *)&mySettings)[i]))
        EEPROM.write(EEPROM_MYSETTINGS_ADDRESS+i, ((unsigned char *)&mySettings)[i]);
  else
    for(unsigned char i=0;i<sizeof(struct networkSettingsStruct);i++)
      ((unsigned char *)&mySettings)[i] = EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+i);
  
}

//Main loop is called over and over again
void loop()
{
  static uint32_t uart_timeout = 0;  //to detect end of packet on bridge serial read
  static unsigned long radioOut_delay = 0;  //delay before write on bridged serial
  static unsigned long resend_to = 0;  //delay before send
  static uint8_t currentSequenceNumber=0;  //currently pending sequence number
  static uint8_t lastAckedSequenceNumber=0;  //last succesfully sent sequence number
  static uint8_t currentpartnerseqnr=0;  //partners sequence number received in last packet
  static uint8_t lastpartnerseqnr=0;  //last partner sequence number we sent an ack for
  static uint8_t retrycnt=0;  //counter for retries
  static uint8_t txRestart=0;  //restart after failed transmit
  
  unsigned char crc, cnt;
  unsigned short i;
  float rssi;
  packet_t * pPacket;  //to access packet information within buffer
  header_t * pHeader;  //to access header information within buffer
  
  //Check for freezes
  wdt_reset();  //At least arduino is running
  
  
/************** Command mode - HWB pulled down ****************/
#define cmdSerial bridge
  
  static unsigned char cmdMode = 0;
  static unsigned long cmdMenuOutput = 0;
  static unsigned char cmdMenuPos = 0;
  static unsigned char settingsStored = 1;
  if (!HWB())
  {
    if (!cmdMode)
    {
      cmdSerial.print("\r\nCmdModeEnter\r\n\r\n");
      cmdMenuOutput=cmdMenuPos=0;
      cmdMode=1;
    }
    for(cnt=cmdSerial.available();cnt||(cnt=cmdSerial.available());cnt--)
    {
#define SETTINGSCOUNT 7
      if (cmdMode&0xF0)
      {
        crc = fromhexch(cmdSerial.read());
        if (!(cmdMode&0xF))
          crc <<= 4;
        switch (cmdMode>>4)
        {
          default: break;
          case 1: mySettings.nwid|=crc;break;
          case 2: mySettings.myid|=crc;break;
          case 3: mySettings.tgtid|=crc;break;
          case 4: mySettings.hopc|=crc;break;
          case 5: mySettings.retries|=crc;break;
        }
        if (cmdMode&0xF)
          cmdMode=1;
        else
          cmdMode++;
      }
      else if (cmdMode==1)
      {
        switch (cmdSerial.read())
        {
          default: break;
          case'n': case'N': cmdMode=0x10; settingsStored = 0; mySettings.nwid=0; break;
          case'm': case'M': cmdMode=0x20; settingsStored = 0; mySettings.myid=0; break;
          case't': case'T': cmdMode=0x30; settingsStored = 0; mySettings.tgtid=0; break;
          case'h': case'H': cmdMode=0x40; settingsStored = 0; mySettings.hopc=0; break;
          case'r': case'R': cmdMode=0x50; settingsStored = 0; mySettings.retries=0; break;
          case'l': case'L': settingsStored = 0; mySettings.accBC^=1; break;
          case's': case'S': 
                settingsStored = 1;
                //Save to EEPROM
                for(unsigned char i=0;i<sizeof(struct networkSettingsStruct);i++)
                {
                  if (EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+i)!=(((unsigned char *)&mySettings)[i]))
                    EEPROM.write(EEPROM_MYSETTINGS_ADDRESS+i, ((unsigned char *)&mySettings)[i]);
                }
                break;
          
          case '+':
            settingsStored = 0;
            switch (cmdMenuPos) {
              default: break;
              case 0: mySettings.nwid++; break;
              case 1: mySettings.myid++; break;
              case 2: mySettings.tgtid++; break;
              case 3: mySettings.hopc++; break;
              case 4: mySettings.retries++; break;
              case 5: mySettings.accBC^=1; break;
              case 6:
                settingsStored = 1;
                //Save to EEPROM
                for(unsigned char i=0;i<sizeof(struct networkSettingsStruct);i++)
                {
                  if (EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+i)!=(((unsigned char *)&mySettings)[i]))
                    EEPROM.write(EEPROM_MYSETTINGS_ADDRESS+i, ((unsigned char *)&mySettings)[i]);
                }
                break;
            } break;
          case '-':
            settingsStored = 0;
            switch (cmdMenuPos) {
              default: break;
              case 0: mySettings.nwid--; break;
              case 1: mySettings.myid--; break;
              case 2: mySettings.tgtid--; break;
              case 3: mySettings.hopc--; break;
              case 4: mySettings.retries--; break;
              case 5: mySettings.accBC^=1; break;
              case 6:
                settingsStored = 1;
                //Save to EEPROM
                for(unsigned char i=0;i<sizeof(struct networkSettingsStruct);i++)
                {
                  if (EEPROM.read(EEPROM_MYSETTINGS_ADDRESS+i)!=(((unsigned char *)&mySettings)[i]))
                    EEPROM.write(EEPROM_MYSETTINGS_ADDRESS+i, ((unsigned char *)&mySettings)[i]);
                }
                break;
            } break;
          case '*':
            cmdMenuPos = (cmdMenuPos+1)%SETTINGSCOUNT; break;
          case '/':
            cmdMenuPos = (cmdMenuPos+SETTINGSCOUNT-1)%SETTINGSCOUNT; break;
        }
      }
      cmdMenuOutput=millis()+20;
    }
    if (cmdMenuOutput<millis())
    {
      cmdSerial.write('\r');
      i = 0;
      
      cmdSerial.print("NID:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(((mySettings.nwid>>4)&0xF),HEX);
      cmdSerial.print(((mySettings.nwid>>0)&0xF),HEX);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print("  MID:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(((mySettings.myid>>4)&0xF),HEX);
      cmdSerial.print(((mySettings.myid>>0)&0xF),HEX);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print("  TID:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(((mySettings.tgtid>>4)&0xF),HEX);
      cmdSerial.print(((mySettings.tgtid>>0)&0xF),HEX);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print("  HOP:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(((mySettings.hopc>>4)&0xF),HEX);
      cmdSerial.print(((mySettings.hopc>>0)&0xF),HEX);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print("  RET:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(((mySettings.retries>>4)&0xF),HEX);
      cmdSerial.print(((mySettings.retries>>0)&0xF),HEX);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print("  LBC:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(mySettings.accBC);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdSerial.print(" STORED:");
      if (cmdMenuPos==i) cmdSerial.write('>');
      cmdSerial.print(settingsStored);
      if (cmdMenuPos==i++) cmdSerial.write('<');
      
      cmdMenuOutput=millis()+500;
    }
    return;
  } else {
    if (cmdMode)
    {
      cmdSerial.print("\r\n\r\nCmdModeLeave\r\n");
      //Reset status variables
      currentSequenceNumber = lastAckedSequenceNumber = currentpartnerseqnr = lastpartnerseqnr = retrycnt = txRestart = 0;
    }
    cmdMode = 0;
  }
  
/************** End command mode *********************/
  
  
  
/************** radio to UART ************************/
  
  //Characterbased output of incoming packets
  /*
  if (radioBuf_available())
  {
    bridge.write(radioBuf_read());
  }
  */
  
  //Buffer based output of incoming packets
  //No chars received since UARTTIMEOUT ms AND have chars in outputbuffer
  //writes bridgeBurst chars every bridgeDelay ms (Let the hardware serial write out async. important for nonblocking RS485 functionality)
  if (uart_timeout<millis() && radioBuf_NE() && radioOut_delay<millis())
  {
    TXEN_ON();
    radioOut_delay = millis()+bridgeDelay;
    for ( i=0 ; i<bridgeBurst && radioBuf_Len()>1 ; i++ )
    {
      bridge.write(radioBuf_read());
    }
    if (radioBuf_Len()==1)  //This is the last char
    {
      bridge.write(radioBuf_read());
      bridge.flush();
      TXEN_OFF();
    }
  }
  
  //CC1101 finished reception of a packet
  if (cnt=wireless.ReadPacket((uint8_t*)&RX_buffer))  //read the packet if any (returns length of packet, 0 if no packet available)
  {
#if DBGLVL>=2
    DBGINFO2(millis()&0xFF,HEX); DBGINFO(" Receive("); DBGINFO(cnt); DBGINFO("): ");
    for (i=0;i<cnt && i<sizeof(header_t)+1;i++) { DBGINFO2(((uint8_t*)&RX_buffer)[i],HEX); DBGWRITE(' '); }
    DBGINFO("-> ");
#endif
    
    //Is the packet length valid?
    if( (RX_buffer.len<sizeof(header_t))||(RX_buffer.len>sizeof(packet_t)) )
        DBGERR("!SIZE\r\n");
    else
    {
      //Set pointers to packet parts
      pPacket = &RX_buffer.packet;
      pHeader = &pPacket->header;
      
      //our network? our ID addressed or (broadcast and we accept boradcasts)?
      if( (pHeader->nwid!=mySettings.nwid) || ((pHeader->dest!=mySettings.myid) && ((pHeader->dest==0xFF)&&(!mySettings.accBC))) )
          DBGERR("!DEST-ID\r\n");
      else
      {
        //Todo actually implement crc check
        cnt = pHeader->crc;
        pHeader->crc = 0;
        crc=42;
        for(i=0 ; i<RX_buffer.len ; i++) crc+=((uint8_t*)pPacket)[i];
        
        //valid packet crc
        if (crc!=cnt)
            DBGERR("!CRC\r\n");
        else
        {
          resend_to = millis()+PACKET_GAP;  //short delay after receive
          
          //No sequencenumber managment on broadcasts
          if (pHeader->dest!=0xFF && pHeader->src!=0xFF)
          {
            currentpartnerseqnr = pHeader->seq;  //remember their sequence number
            if (currentSequenceNumber == pHeader->pseq) ERRLEDOFF();  //got ack for our last data
            lastAckedSequenceNumber = pHeader->pseq;  //remember last acked sequence number
          }
          
#if DBGLVL>=2
          //Calculate RSSI
          crc = pPacket->data[pHeader->len]; rssi = crc; if (crc&0x80) rssi -= 256; rssi /= 2; rssi -= 74;
          DBGINFO(" RSSI: "); DBGINFO(rssi); DBGINFO("dBm");
#endif
          
          //did they alter their sequence number? implies they got an ack and sent new data
          if (pHeader->dest==0xFF || pHeader->src==0xFF || currentpartnerseqnr!=lastpartnerseqnr)
          {
            for(i=0 ; i<pHeader->len && radioBuf_Free() ; i++)  //fill uart output buffer
            {
              radioBuf_write(pPacket->data[i]);
            }
            //If buffer is filled beyond threshold
            if (radioBuf_Len()>=RADIO_SENDTHRES && pHeader->dest!=0xFF && pHeader->src!=0xFF)
              resend_to += PACKET_PAUSE;  //delayed ack, allows us to drain some buffer content
          }
          //got packet with same sequence number we already acked -> retransmit ack
          else if (pHeader->len && currentpartnerseqnr==lastpartnerseqnr)
          {
            lastpartnerseqnr--;
          }
          
#if DBGLVL>=2
          DBGINFO(" buf: "); DBGINFO(radioBuf_Len()); DBGINFO("\r\n");
#endif
        }
      }
    }
  }
  
/************** UART to radio ************************/
  
  //read incoming chars from UART
  //still have bufferspace and remaining chars on uart
  for (cnt=bridge.available() ; uartBuf_NF() && ( cnt  || ((uartBuf_Len()<=UART_SENDTHRES) && (cnt=bridge.available()))) ; cnt--)
  {
    uartBuf_write(bridge.read());
    uart_timeout = millis()+UART_TIMEOUT;
  }
  
  //delay between packets is over
  if (resend_to<millis())
  {
    //Last send not finished yet
    if (wireless.TXBusy())
    {
      DBGINFO("Still sending\r\n");
      resend_to= millis()+2;
    }
    //(restart last transmit OR have Data AND datablock on uart complete) OR buffer over threshold OR waiting for ack OR partner sent new data
    else if (txRestart || ((uartBuf_NE() && uart_timeout<millis()) || uartBuf_Len()>=UART_SENDTHRES || currentSequenceNumber!=lastAckedSequenceNumber || currentpartnerseqnr!=lastpartnerseqnr))
    {
      DBGINFO2(millis()&0xFF,HEX); DBGINFO(" Buf:"); DBGINFO(uartBuf_Len()); DBGINFO(" ");
      
      //Init packet with common values
      pPacket = &TX_packet;
      pHeader = &pPacket->header;
      pHeader->nwid = mySettings.nwid;
      pHeader->src = mySettings.myid;
      pHeader->dest = mySettings.tgtid;
      pHeader->pseq = currentpartnerseqnr;
      pHeader->hopc = mySettings.hopc;
      
      //Broadcasts are not acked so update status as if received one
      if (pHeader->src==0xFF || pHeader->dest==0xFF)
      {
        lastpartnerseqnr=currentpartnerseqnr;
        lastAckedSequenceNumber=currentSequenceNumber;
      }
      
      //forced retry caused by transmit error
      if (txRestart)
      {
        ERRLEDON();
        DBGINFO("Repeat (");
        lastpartnerseqnr = currentpartnerseqnr;  //update ack signal
      }
      
      //still no ack for last sent sequence number
      if (currentSequenceNumber!=lastAckedSequenceNumber)
      {
        //retried too often
        if (retrycnt>=mySettings.retries)
        {
          ERRLEDON();
          lastAckedSequenceNumber = currentSequenceNumber;  //start next transmission with a different sequence number
          pHeader->seq = currentSequenceNumber;
          retrycnt = 0;  //reset retries
          uartBuf_Clear();  //kill buffer
          DBGERR("reached max Retries\r\n");
          return;
        } else {
          ERRLEDON();
          DBGINFO("Retry #"); DBGINFO(retrycnt); DBGINFO(" (");
          lastpartnerseqnr = currentpartnerseqnr;  //update ack signal
          retrycnt++;
        }
      }
      
      //all our data was acked
      else
      {
        lastpartnerseqnr = currentpartnerseqnr;  //update ack signal
        
        //Next data chunk [ (we have data AND uart timed out) OR uart beyond threshold ]
        if ((uartBuf_NE() && uart_timeout<millis()) || uartBuf_Len()>=UART_SENDTHRES)
        {
          //Refill packet buffer with new data
          do {
            pHeader->seq = ++currentSequenceNumber;  //new data -> increase sequence number
          } while (!currentSequenceNumber);  //0 is reserved for "not used before"
          
          //No ACK on broadcasts -> fake one
          if (mySettings.myid==0xFF || mySettings.tgtid==0xFF)
              lastAckedSequenceNumber=currentSequenceNumber;
          
          pHeader->len = uartBuf_Len()<MAXDATALEN ? uartBuf_Len() : MAXDATALEN;  //amount of data to be transmitted
          for ( i=0 ; i<pHeader->len ; i++ )
          {
            pPacket->data[i] = uartBuf_read();
          }
          
          DBGINFO("Sending (");
          retrycnt = 0;
        }
        
        //no new data to send -> send only ack
        else
        {
          //no ACKs on broadcasts
          if (mySettings.myid==0xFF || mySettings.tgtid==0xFF)
          {
            DBGINFO("NoAck on BC\r\n");
            return;
          }
          DBGINFO("SendAck (");
          pHeader->len = 0;
        }
      }
      
      //TODO calculate real crc
      pHeader->crc=0;
      crc=42;
      for( i=0 ; i<(sizeof(header_t)+pHeader->len) ; i++)
      {
        crc+=((uint8_t*)&TX_packet)[i];
      }
      pHeader->crc=crc;
      
      DBGINFO(i); DBGINFO(","); DBGINFO(pHeader->len); DBGINFO(")\r\n");
      
      //Transmit packet
      if (wireless.SendPacket((uint8_t*)&TX_packet,i))
      {
        if (txRestart) ERRLEDOFF();
        txRestart = 0;
        //Smaller delay on boradcasts because there is no ACK
        if (pHeader->src==0xFF || pHeader->dest==0xFF)
          resend_to = millis()+PACKET_GAP_NOACK;
        else
          resend_to = millis()+RESEND_TIMEOUT;
      } else {
        //transmit error .. retry later
        txRestart = (txRestart) ? (txRestart-1) : SEND_RETRIES_ON_ERROR;
        resend_to = millis()+PACKET_GAP_ERR;
        DBGERR("ERROR Transmit\r\n");
      }
    }
  }
}

